/* pluralizer.c - make plural english nouns. */
/* Created: Sun May 16 10:07:03 PM EDT 2021 rahjiii */
/* Copyright © 1991-2021 The Last Outpost Project */
/* $Id: pluralizer.c,v 1.1 2021/05/19 13:58:24 malakai Exp $ */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <gmodule.h>	/* glib header for hash table module */

/* these three includes are needed for lo_log and lo_debug */
#include "structs.h"
#include "utils.h"
#include "debug.h"

#include "scnprintf.h"
#include "pluralizer.h"

#ifndef MAX_STRING_LENGTH
#define MAX_STRING_LENGTH 4096
#endif

/* change this to whatever works for you.  Should be found in current working
 * directory at time pluralize_init() is called. */
#define NOUNS_S2P_FILENAME "plural_nouns.txt"


/* local variables */
GHashTable *nouns_s2p = NULL;	/* hash table dictionary will be created by pluralize_init() */

/* init the pluralizer noun singular to plural database. Call this once
 * somewhere early in your boot up routine.  If you call the other functions
 * without calling this init, you won't have irregualr pluralization lookups.*/
int pluralize_init(void) {

	FILE *nin;
	char singular[MAX_STRING_LENGTH];
	char plural[MAX_STRING_LENGTH];
	char *c;

	if(nouns_s2p) {
		/* already loaded. */
		return(1);
	}

	/* load up the nouns singular to plural dictionary */

	if( !(nin = fopen(NOUNS_S2P_FILENAME,"r"))) {
		lo_log("%s: %s",NOUNS_S2P_FILENAME,strerror(errno));
		return(0);
	}

	nouns_s2p = g_hash_table_new(g_str_hash, g_str_equal);
	while( (fscanf(nin,"%s %s",singular,plural) == 2) ) {
		for(c=singular;*c;*c++=tolower(*c));
		for(c=plural;*c;*c++=tolower(*c));
		lo_debug(DEBUG_DB,"Adding '%s'->'%s'",singular,plural);
		g_hash_table_insert(nouns_s2p,g_strdup(singular),g_strdup(plural));
	}
	lo_debug(DEBUG_DB,"loaded %d singular to plural nouns.",
		g_hash_table_size(nouns_s2p)
	);

	fclose(nin);
	return(1);

}

/* Call this on shutdown, to free up the hash dictionary. */
void pluralize_free(void) {
	if(nouns_s2p) {
		g_hash_table_destroy(nouns_s2p);
	}
}


/* does what it says on the tin. */
int is_vowel(char c) {
	switch (tolower(c)) {
	case 'a':
	case 'e':
	case 'i':
	case 'o':
	case 'u':
	case 'y':
		return(1);
	default:
		return(0);
	}
}

/* spells out a number between -99 and 99.  Anything outside that range is
 * printed numerically.  Free whatever string this gives back to you. */

char *int_to_words(int num) {

	char *ones_str[] = {
		"zero",
		"one",
		"two",
		"three",
		"four",
		"five",
		"six",
		"seven",
		"eight",
		"nine",
		"ten",
		"eleven",
		"twelve",
		"thirteen",
		"fourteen",
		"fifteen",
		"sixteen",
		"seventeen",
		"eighteen",
		"nineteen"
	};

	char *tens_str[] = {
		"zero",
		"ten",
		"twenty",
		"thirty",
		"forty",
		"fifty",
		"sixty",
		"seventy",
		"eighty",
		"ninety"
	};

	char buf[MAX_STRING_LENGTH];

	int ones, tens;
	char *posneg;
	char positive[] = "";
	char negative_word[] = "minus ";
	char negative_sym[] = "-";
	int isneg=0;
	if(num < 0) {
		isneg = 1;
		num = -num;
	}
	posneg = positive;

	if(num>99) {
		if(isneg)
			posneg = negative_sym;
		scnprintf(buf,sizeof(buf),"%s%d",posneg,num);
	} else {
		if(isneg)
			posneg = negative_word;
		if(num < 20) {
			scnprintf(buf,sizeof(buf),"%s%s",posneg,ones_str[num]);
		} else {
			tens = num/10;
			ones = num%10;
			if(ones == 0) {
				scnprintf(buf,sizeof(buf),"%s%s",posneg,tens_str[tens]);
			} else {
				scnprintf(buf,sizeof(buf),"%s%s-%s",posneg,tens_str[tens],ones_str[ones]);
			}
		}
	}
	return(strdup(buf));
}

/* Pluralize a singular english word.  ex given "sword" returns "swords".  ONLY
 * works on single words, with no spaces.  Based on rules listed at
 * https://www.grammarly.com/blog/plural-nouns/ Be sure to free the string that
 * this gives back to you.*/

char *pluralize_noun(char *singular,int count) {

	int slen;
	char *s,*send, *s1,*s2;
	int r1,r2;
	char plural[MAX_STRING_LENGTH];
	char *p,*pend;
	char *word,*w;

	if(!singular) return(NULL); /* GIGO. */
	
	s = singular;
	slen = strlen(singular);
	send = s+slen;

	p = plural;
	pend = plural + sizeof(plural);

	/* if there's exactly 1, there's nothing to do. */
	if(count==1) {
		return(strdup(singular));
	}

	/*irregular pluralizaation dictionary lookup first. */
	if(nouns_s2p) {
		word = strdup(singular);
		for(w=word;*w;*w++=tolower(*w)); /*downcase a copy of the word.*/
		w = g_hash_table_lookup(nouns_s2p,word);
		free(word);
		if(w) {
			return(strdup(w));
		}
	}

	/* ok... derive the plural from some rules. */

	r1 = MAX(slen-1,0); /* root-1 of singular */
	r2 = MAX(slen-2,0); /* root-2 of singular */
	s1 = singular + r1; /* last char of word */
	s2 = singular + r2; /* last two chars of word */

	/* RULE: If the singular noun ends in -ss, -sh, -ch, -x, or -z, add ‑es
	 * to the end to make it plural.  (In some cases, singular nouns ending in
	 * -s or -z, require that you double the -s or -z prior to adding the -es
	 * for pluralization, but skipping that for now.) */
	if( !strcasecmp(s2,"ss") ||
		!strcasecmp(s2,"sh") ||
		!strcasecmp(s2,"ch") ||
		!strcasecmp(s1,"x") ||
		!strcasecmp(s1,"o") ||
		!strcasecmp(s1,"z")
	) {
		scnprintf(p,pend-p,"%ses",singular);
		return(strdup(plural));
	}

	/* RULE: If the noun ends with ‑f or ‑fe, the f is often changed to ‑ve before
	 * adding the -s to form the plural version. */
	if( !strcasecmp(s2,"fe") ) {
		scnprintf(p,pend-p,"%.*sves",r2,singular);
		return(strdup(plural));
	}
	if( !strcasecmp(s1,"f") &&
		strcasecmp(s2,"ff")  /* if it ends if ff, leave it alone. */
	) {
		scnprintf(p,pend-p,"%.*sves",r1,singular);
		return(strdup(plural));
	}

	/* RULE: If the singular noun ends in -y... */
	if(!strcasecmp(s1,"y")) {
		if(is_vowel(*s2)) {
			/* ...and the letter before the -y is a vowel, simply add an -s to
			 * make it plural. */
			scnprintf(p,pend-p,"%ss",singular);
			return(strdup(plural));
		} else {
			/* ...and the letter before the -y is a consonant, change the
			 * ending to ‑ies to make the noun plural. */
			scnprintf(p,pend-p,"%.*sies",r1,singular);
			return(strdup(plural));
		}
	}

	/* RULE: If the singular noun ends in ‑us, the plural ending is frequently ‑i. */
	if( !strcasecmp(s2,"us") ) {
		scnprintf(p,pend-p,"%.*si",r2,singular);
		return(strdup(plural));
	}

	/* RULE: If the singular noun ends in ‑is, the plural ending is changed to -es. */
	if( !strcasecmp(s2,"is") ) {
		scnprintf(p,pend-p,"%.*ses",r2,singular);
		return(strdup(plural));
	}

	/* RULE: If the singular noun ends in ‑on, the plural ending is ‑a. 
	if( !strcasecmp(s2,"on") ) {
		snprintf(p,pend-p,"%.*sa",r2,singular);
		return(strdup(plural));
	}
	*/

	/* RULE: If the singular noun ends in ‑s, the plural ending is -es. */
	if( !strcasecmp(s1,"s") ) {
		scnprintf(p,pend-p,"%ses",singular);
		return(strdup(plural));
	}
	
	/* RULE: To make regular nouns plural, add ‑s to the end. */
	scnprintf(p,pend-p,"%ss",singular); 
	return(strdup(plural)); 
}

/* pluralize a singlular english nown phrase.  Turns a phrase like "a short
 * sword" into "six short swords" If count is zero, returns pluralized word
 * without a count.  This code uses heuristics and may sometimes produce
 * laughable crap.  Be sure to free the string that this gives back to you.*/

char *pluralize_noun_phrase(char *singular, int count) {

	int slen;
	char *s,*send;

	char plural[MAX_STRING_LENGTH];
	char *p,*pend;

	char *pluralized;
	char *lastblank;

	char *written_count;
	
	s = singular;
	slen = strlen(singular);
	send = s+slen;

	p = plural;
	pend = plural + sizeof(plural);

	/* if there is an article prefix of a, an, the, one, skip over it.*/
	for (;*s && isblank(*s);*p++ = *s++); /* copy whitespace to start of next word. */
	if( !strncasecmp(s,"a ",2) ) {
		s+=2;
	}
	if( !strncasecmp(s,"an ",3) ) {
		s+=3;
	}
	if( !strncasecmp(s,"the ",4) ) {
		s+=4;
	}
	if( !strncasecmp(s,"one ",4) ) {
		s+=4;
	}

	/* if count >=0, write the count to the result. */
	if(count >=0) {
		written_count = int_to_words(count);
		p+=scnprintf(p,pend-p,"%s ",written_count);
		free(written_count);
	}

	/* copy into the result, up to end of string or the preposition " of ", so
	 * that noun prhases like "a bag of holding" are pluralized as bags, not
	 * holdings. */
	for(lastblank=p;*s && strncasecmp(s," of ",4) && p<pend; *p++ = *s++) {
		if(isblank(*s)) {
			lastblank = p+1;
		}
	}
	*p='\0';

	/* back up, find the space, and pluralize that last word. */
	p = lastblank;
	pluralized = pluralize_noun(p,count);
	p+=scnprintf(p,pend-p,"%s",pluralized);
	free(pluralized);

	/* copy the rest of the line into the result unchanged. */
	for(;*s && p<pend; *p++ = *s++);
	*p='\0';
	return(strdup(plural));

}
